/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.loadstore.c,v 1.4 1999/07/15 23:34:37 muth Exp $
  Comment:  This file contains the optimization code for
            Load Avoidance
            Store Avoidance
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsValueAvailibilityInExtBbl"

extern REGLIST AinsValueAvailibilityInExtBbl(INDEX ains, REGLIST avail, INDEX bins, INDEX ibbl)
{
    INDEX iins;

    if( AINS_BBL(ains) != ibbl )
    {
        CASSERT( AbblNpreds(ibbl) == 1,("weird\n"));
        avail = AinsValueAvailibilityInExtBbl(ains,avail,bins, AEDG_HEAD(ABBL_PRED(ibbl)));
        iins = ABBL_INS_FIRST(ibbl);
    }
    else
    {
        iins = AINS_NEXT(ains);
    }
    
    for( ; VALID(iins) && iins!=bins; iins = AINS_NEXT(iins) )
    {
        if( AINS_CODE(iins) == I_lda && AINS_DATA(iins) == 0 && (avail & AinsUse(iins)) )
        {
            avail |= AinsDef(iins);
        }
        else if( AINS_CODE(iins) == I_cpys && avail & AinsUse(iins) && AINS_REG_A(iins) == AINS_REG_B(iins) )
        {
            avail |= AinsDef(iins);
        }
        else
        {
            avail &= ~AinsDef(iins);
        }
    }
    
    return avail;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "RegFixupLivenessInExtendedBbl"

static void RegFixupLivenessInExtendedBbl(REG reg, INDEX ibbl, INDEX jbbl)
{
    while( jbbl != ibbl )
    {
        CASSERT( AbblNpreds(jbbl) == 1,("weird\n"));
        jbbl =  AEDG_HEAD(ABBL_PRED(jbbl));
        ABBL_OUT(jbbl) |= REG2BIT(reg);
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "DataTypesOk"

static BOOL DataTypesOk(DATATYPE t1, DATATYPE t2)
{
    if( t1 == DT_QUAD && t2== DT_QUAD ) return 1;
    if( t1 == DT_LONG && t2== DT_LONG ) return 1;
    if( t1 == DT_GFLT && t2== DT_GFLT ) return 1;
    if( t1 == DT_TFLT && t2== DT_TFLT ) return 1;
    if( t1 == DT_QUAD && t2 == DT_LONG ) return 1;
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindDefiningLoadStoreInExtBbl"

static INDEX FindDefiningLoadStoreInExtBbl(INDEX load)
{
    INDEX iins;
    INDEX ibbl = AINS_BBL(load);

    iins = AINS_PREV(load);
    
    for(;;)
    {
        for( ; VALID(iins); iins = AINS_PREV(iins) )    
        {
            NUMBER result;

            if( AINS_IS_STORE(iins) )
            {
                result = DoMemoryAccessesConflict(iins,load);
                
                if( result == CONFLICT_MAYBE )
                    return INVALID;
                
                if( result == CONFLICT_NO )
                    continue;
                
                if( result == CONFLICT_MATCH )
                {
                    /* WARNING(("Potential\n")); */
                    return INVALID;
                }
                
                if( result == CONFLICT_EXACT )
                {
                    if( DataTypesOk(AinsDataType(iins),AinsDataType(load)) )
                        return iins;
                    else
                        return INVALID;
                }
                
                ERROR(("unexpected aliasing result store: %d\n",result));
            }
            else if( AINS_TYPE(iins)==IT_ILD || AINS_TYPE(iins) == IT_FLD )
            {
                result = DoMemoryAccessesConflict(iins,load);
                
                if( result == CONFLICT_MAYBE )
                    continue;
                                
                if( result == CONFLICT_MATCH )
                {
                    /* WARNING(("Potential\n")); */
                    continue;
                }
                
                if( result == CONFLICT_NO )
                    continue;
                
                if( result == CONFLICT_EXACT )
                {
                    if( DataTypesOk(AinsDataType(iins),AinsDataType(load)) )
                        return iins;
                    else
                        continue;
                }
                
                ERROR(("unexpected aliasing result load: %d\n",result));
            }
        

        }
        if( AbblNpreds(ibbl) != 1 ) return INVALID;

        ibbl = AEDG_HEAD(ABBL_PRED(ibbl));
        iins = ABBL_INS_LAST(ibbl);
    }
}

/*=======================================================
@OPT
Avoid reloading of a value from memory.
Pairs of store/load instruction that provably reference the same memory
location are replace by a store/move pair.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptAvoidLoads"

extern BOOL OptAvoidLoads(FILE *fp, BOOL no_code_growth)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Avoidloads");
    BOOL   change;

    REGLIST avail;
    
    REG    reg_src;
    REG    reg_dst;

    INDEX load;
    INDEX iins;
    INDEX ifun;
    INDEX ibbl;
    
    NUMBER lfcount = 0;
    NUMBER lfcounth = 0;
    NUMBER lpcount = 0;
    NUMBER lpcounth = 0;

    NUMBER sfcount = 0;
    NUMBER sfcounth = 0;
    NUMBER spcount = 0;
    NUMBER spcounth = 0;
    
    NUMBER bcount = 0;
    NUMBER bcounth = 0;

    INT64 weight = 0;

    clock_t time = clock();

    FOREACH_LIVE_FUN(ifun)
    {        
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            FOREACH_BBL_INS(ibbl,load)
            {
                if( AINS_TYPE(load) != IT_ILD && AINS_TYPE(load) != IT_FLD  ) continue;

                iins = FindDefiningLoadStoreInExtBbl(load);
                
                if( AINS_TYPE(iins) == IT_IST || AINS_TYPE(iins) == IT_FST )
                {
                    reg_src = AINS_REG_A(iins);

                    if( (AINS_FLAGS(iins) & FI_EVAL_A) && AINS_A_EVAL(iins) != 0 )
                    {
                        POTENTIAL(("Constant forwarding value: %016lx  reg %d  (%ld)\n",
                                 AINS_A_EVAL(iins),reg_src,ABBL_COUNT(AINS_BBL(iins))));
                    }
                    
                    spcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        spcounth++;
                }
                else if( AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD )
                {
                    reg_src = AINS_REG_C(iins);
                    lpcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        lpcounth++;

                }
                else continue;
                

                reg_dst = AINS_REG_C(load);

                if( verbose )
                {
#if 0
                    fprintf(fp,"\t[*] Load Forwarding Opportunity %d (%d) -> %d (%d):\n",iins,AINS_BBL(iins),load, AINS_BBL(load));
                    fprintf(fp,"\t[*] FROM: ");
                    PrintIns(fp,iins);
                    fprintf(fp,"\t[*] TO  : ");
                    PrintIns(fp,load);
#endif
#if 0
                    PrintBblSuper(stdout,AINS_BBL(iins));
                    PrintBblSuper(stdout,ibbl);
#endif
                }


                avail = AinsValueAvailibilityInExtBbl(iins,REG2BIT(reg_src),load,ibbl);

                avail &= REG_IS_INT(reg_dst) ? REGS_INT : REGS_FLT;
                         
                if( avail )
                {
                    reg_src = RegsGetFirst(avail);
                    AinsMakeMove(load,reg_src,AinsDataType(iins),reg_dst,AinsDataType(load));
                    RegFixupLivenessInExtendedBbl(reg_src, AINS_BBL(iins),AINS_BBL(load));                    
                    goto LabelDone;
                }

                if( no_code_growth )
                    continue;

                /* we cannot forward it directly, try to use a bypass register */

                avail = MyFindScratchRegsInExtBbl(iins,load);
                                    
                avail &= REG_IS_INT(reg_dst) ? REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

                if( avail )
                {
                    INDEX iins_bypass = AinsNew();
                    INDEX jbbl;
                    REG   reg_bypass;

                    reg_bypass = RegsGetFirst(avail);

                    jbbl = ibbl;
                    while( jbbl != AINS_BBL(iins) )
                    {
                        jbbl = AEDG_HEAD(ABBL_PRED(jbbl));
                        ABBL_OUT(jbbl) |= REG2BIT(reg_bypass);
                    }
                    
                               
                    AINS_BBL(iins_bypass) = AINS_BBL(iins);
                    AinsMakeMove(iins_bypass,reg_src,AinsDataType(iins),reg_bypass,AinsDataType(iins));
                    AinsInsertAfter(iins_bypass,iins);
                    RegFixupLivenessInExtendedBbl(reg_bypass, AINS_BBL(iins_bypass),AINS_BBL(load));                    
                    AinsMakeMove(load, reg_bypass,AinsDataType(iins),reg_dst,AinsDataType(load));
                    
                    bcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        bcounth++;
                    goto LabelDone;
                }

                if( verbose )
                {
#if 0
                    fprintf(fp,"\t[*] Failed\n");
#endif
                }
                
                continue;

            LabelDone:
                
                if( verbose )
                {


                    fprintf(fp,"\t[*] Load Forwarded %d (%d) -> %d (%d):\n",iins,AINS_BBL(iins),load, AINS_BBL(load));
                    fprintf(fp,"\t[*] FROM: ");
                    PrintIns(fp,iins);
                    fprintf(fp,"\t[*] TO : ");
                    PrintIns(fp,load);
                    if( ABBL_FUN(AINS_BBL(iins)) != ABBL_FUN(AINS_BBL(load)) )
                    {
                        HURRAH(("Interproc\n"));
                        
                    }                    
#if 0
                    PrintBblSuper(stdout,AINS_BBL(iins));
                    PrintBblSuper(stdout,AINS_BBL(load));

                    if( ABBL_FUN(AINS_BBL(iins)) != ABBL_FUN(AINS_BBL(load)) )
                    {
                        PrintFunSuper(stdout, ABBL_FUN(AINS_BBL(iins)));
                        PrintFunSuper(stdout, ABBL_FUN(AINS_BBL(load)));
                        
                    }
                    
#endif
                }

                if( AINS_TYPE(iins) == IT_IST || AINS_TYPE(iins) == IT_FST )
                {
                    sfcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        sfcounth++;
                }
                else
                {
                    lfcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        lfcounth++;
                }

                weight += ABBL_COUNT(AINS_BBL(load));
            }
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = sfcount+lfcount>0;

    InfoStatus(fp,"\t[OptAvoidloads%c] SFor %d#%d  SPot: %d#%d LFor: %d#%d  LPot: %d#%d  BPass: %d#%d  wt: %ld {%d ms}\n",
            CHANGE(change),sfcount,sfcounth,spcount,spcounth,lfcount,lfcounth,lpcount,lpcounth,bcount,bcounth,weight,time);
    
    return change;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindFollowupStore"

static INDEX FindFollowupStore(INDEX store)
{
    INDEX iins;
    INDEX ibbl = AINS_BBL(store);

    iins = AINS_NEXT(store);
    
    for(;;)
    {
        for( ; VALID(iins); iins = AINS_NEXT(iins) )    
        {
            NUMBER result;

            if( AINS_IS_STORE(iins) )
            {
                result = DoMemoryAccessesConflict(iins,store);
                
                if( result == CONFLICT_EXACT )
                {
                    return iins;
                }
                else
                {
                    continue;
                }
                
            }
            else if( AINS_IS_LOAD(iins) )
            {
                result = DoMemoryAccessesConflict(iins,store);
                
                if( result == CONFLICT_NO )
                {
                    continue;
                }
                else
                {
                    return INVALID;
                }
                
            }
        

        }
        
        if( AbblNsuccs(ibbl) != 1 ) return INVALID;

        ibbl = AEDG_TAIL(ABBL_SUCC(ibbl));
        iins = ABBL_INS_FIRST(ibbl);
    }
}

/*=======================================================
@OPT

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptAvoidStores"

extern BOOL OptAvoidStores(FILE *fp, BOOL no_code_growth)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Avoidstores");
    BOOL   change;

    INDEX store;
    INDEX iins;
    INDEX ifun;
    INDEX ibbl;
    
    NUMBER sfcount = 0;
    NUMBER sfcounth = 0;
    NUMBER sicount = 0;
    NUMBER sicounth = 0;

    INT64 iweight = 0;
    INT64 fweight = 0;

    clock_t time = clock();

    FOREACH_LIVE_FUN(ifun)
    {
        FOREACH_FUN_BBL(ifun,ibbl)
            FOREACH_BBL_INS(ibbl,store)
            {
                if( AINS_TYPE(store) != IT_IST && AINS_TYPE(store) != IT_FST ) continue;
                
                iins = FindFollowupStore(store);

                if( !VALID(iins) ) continue;

                if( AINS_TYPE(iins) != IT_IST && AINS_TYPE(iins) != IT_FST ) continue;

                if( AinsMemWidth(store) > AinsMemWidth(iins) )
                {
#if 0
                    WARNING(("strange case\n"));
                    fprintf(fp,"\t[*] store %d -> store %d\n",store,iins);
                    PrintBblSuper(stdout,ibbl);
                    if( ibbl != AINS_BBL(iins) )
                        PrintBblSuper(stdout,AINS_BBL(iins));
#endif
                    continue;
                }
                
                if( verbose )
                {
                    fprintf(fp,"\t[*] store %d -> store %d\n",store,iins);
                    PrintBblSuper(stdout,ibbl);
                    if( ibbl != AINS_BBL(iins) )
                        PrintBblSuper(stdout,AINS_BBL(iins));
                }
                
                if( AINS_TYPE(store) == IT_IST )
                {
                    iweight += ABBL_COUNT(ibbl);
                    sicount++;
                    if( ABBL_IS_HOT(ibbl) ) sicounth++;
                    AinsKill(store);
                }
                else if( AINS_TYPE(store) == IT_FST )
                {
                    fweight += ABBL_COUNT(ibbl);
                    sfcount++;
                    if( ABBL_IS_HOT(ibbl) ) sfcounth++;
                    AinsKill(store);
                }
                else
                {
                    ERROR(("unexpected case\n"));
                }
            }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = sfcount+sicount>0;

    InfoStatus(fp,"\t[OptAvoidstores%c] int stores killed %d#%d  wt: %ld   flt stores killed: %d#%d   wt: %ld {%d ms}\n",
            CHANGE(change),sicount,sicounth,iweight,sfcount,sfcounth,fweight,time);
    
    return change;
}

/*=======================================================
=======================================================*/
      
